// code by jph

// distances in meter, angles in rad, time in seconds

import java.io.*;

public class Marsokhod {
  static final String[]       MID={"FL","FR","CL","CR","BL","BR","AF","AR"};
  static final double         MLA=0.413;                // distance from middle axis to active joint axis
  static final double         LEV=0.145;                // active joint to exterior axis
  static final double         TLT=0.524;                // max tilt of axles    
  static final double         VEL=2.5;                  // maximum vel [rad / sec] at 24.0265 _VCC is 2.8274
  static final double         ANG=Math.acos(LEV/MLA);   // + and - inclination range of arm joints
  static final double         EXP=1.3;                  // interval size for bending joints
  static final double         TOL=0.2;                  // tolerance for bending joints
  static final double[]       BST={0.9896,0.9808,0.9703,0.9412,0.9713,0.9469,0.8500,0.8500};  // 1->6, 5->7 boost for interval logic  
  static final double[]       BCL={0,0,0,0,0,0,0.0524,0.2500};            // backlash FRONT BACK
  //static final double[]       LOD={0.1960,0.5592,0.2449};                 // con=1 32 [Hz]  [Ringbuffer 6 confirmed on July 24th]
  static final double[]       LOD={0.1008,0.2798,0.3742,0.2452};                 // con=1 32 [Hz]  [Ringbuffer 6 confirmed on July 24th]
  static final double[]       TLA={625 , 412};
  static final double[]       TLB={444 , 624};
  static final double[][]     ACC={{145,413},
                                   {120,383},
                                   { 75,376}};
  static RingBuffer<Double>   VEX=new RingBuffer<Double>(8*6,new Double(0)); // [8*6 for con=1 as confirmed on July 24th]
  static final double[]       XIS={0.70,0.47};          // {0.54,0.36};    // factor for PWM, LOD^2
  static volatile double      UPS=1.5;
  static volatile double      RAD=0.12;                 // wheel radius INDOOR=>0.125 OUTDOOR=>0.115
  static volatile int[]       con=new int[8];
  static volatile int[]       icr=new int[8];
  static volatile int         msk=0xff;                 // indicates which of the wheels are connected
  
  static volatile double[]    bot={Double.NEGATIVE_INFINITY,Double.NEGATIVE_INFINITY,Double.NEGATIVE_INFINITY,Double.NEGATIVE_INFINITY,Double.NEGATIVE_INFINITY,Double.NEGATIVE_INFINITY,-ANG-EXP,ANG    };  // BOT is used for control with bounded position, values change adaptively
  static volatile double[]    top={Double.POSITIVE_INFINITY,Double.POSITIVE_INFINITY,Double.POSITIVE_INFINITY,Double.POSITIVE_INFINITY,Double.POSITIVE_INFINITY,Double.POSITIVE_INFINITY,-ANG    ,ANG+EXP};  // TOP is used for control with bounded position, values change adaptively
  
  static volatile long[]      enc=new long[8];	        // encoder ticks
  static volatile double[]    pos=new double[8];        // position [rad]
  static volatile double[]    vel=new double[8];        // velocity [rad / sec]
  static volatile double[]    vex=new double[8];        // no load velocity [rad / sec]
  static volatile double[]    lod=new double[8];        // load [rad / sec]
  static volatile int[]       pwm=new int[8];           // but read write
  static volatile double[]    bcl=new double[8];        // true angle of wheel/joint, purpose is to compensate backlash
  
  static volatile double      YAW=509;                  // estimation of bias-offset
  static volatile int         yaw=509;                  // raw gyro adc value
  static volatile double      gyr=0;                    // gyro rate [rad / sec]
  
  static volatile double[]    acc=new double[3];        // raw acceleration from sensor
  static volatile double[]    tlt=new double[2];        // tilt of exterior axles with respect to center axle
  
  static final double         _VCC=24.0265;             // calibration of wheels at _VCC
  static volatile double      VCC=_VCC;                 // voltage measured at the power distributing board when robot is stopped
  static volatile double      vcc=_VCC;                 // voltage measured at the power distributing board
  
  static volatile double[]    grv=new double[3];        // estimation of direction to center of Earth
  
  static final int            LED_GREEN=1;
  static final int            LED_WHITE=2;
  static final int            LED_CALIB=4;
  static volatile int         led=0;
  
  static volatile double      lod_mea=0;                // mean of load
  static volatile double      lod_var=0;                // variance of load
  
  static final int            FLD=6;
  static volatile double[][]  fld=new double[FLD][8];
  static volatile double[][]  fle=new double[FLD][8];
  
  static volatile boolean     dirty=false;              // indicates if configuration frame should be sent
  
  static final int            LOGIC_CALIBRATION   =1;
  static final int            LOGIC_SMART_PLACE   =2;
  static final int            LOGIC_SLOPE_RESPONSE=4;
  static final int            LOGIC_TURN_CONTRACT =8;
  static final int            LOGIC_LOAD_DISTRIB_V=16;
  static final int            LOGIC_LOAD_DISTRIB_S=32;
  static volatile int         logic=0;

  static final int            CALIB_PASSIVE=0;
  static final int            CALIB_ELEF_UP=1;
  static final int            CALIB_ELEF_DN=2;
  static final int            CALIB_ELEB_UP=3;
  static final int            CALIB_ELEB_DN=4;
  static final int            CALIB_IGNITE =5;
  static volatile int         calib=CALIB_PASSIVE;
  static volatile double      cal=0;
  static final double         CAL=0.25;                 // joint speed during calibration
  static volatile double      fuse=0;                   // this becomes 1 when ignited
  
  static volatile double[]    rat=new double[8];        // control velocity at no load, values are in interval (-2.84,...,2.84)
  static volatile double[]    ral=new double[8];        // redistribution due to load
  
  static final int            PUSH=0;                   // drive straight
  static final int            TURN=1;                   // spin around the center of the center axle
  static final int            DIGG=2;                   // digg with front right
  static final int            DRFT=3;                   // drift sidewards
  static final int            ELEF=4;                   // bend front joint
  static final int            ELEB=5;                   // bend rear joint
  
  static Control[]            myControl=new Control[FLD];
  static volatile double      thr=1;                    // throttle
  static volatile double      kap=1;                    // truncate the control commands if any VEL<rat
  static volatile double      joi=0.3;                  // load threshold to protect arms
  
  static IntervalMap          myIntervalMap;            // V_icr(d)
  
  static Rolk[]               myRolk=new Rolk[5];
  static final int            ROLK_FRONT =0;
  static final int            ROLK_FULL  =1;
  static final int            ROLK_REAR  =2;
  static final int            ROLK_WORM  =3;
  static final int            ROLK_EXTEND=4;  
  static final String[]       ROLK={"front","full","rear","worm","extend"};
  
  
  static void init() {
    int c1;
	  for (c1=0;c1<8;c1++) {
  	  icr[c1]=511;
  	  con[c1]=0;
	  }
	  
    myIntervalMap=new IntervalMap(1000,0);
    myIntervalMap.Doubles("res/vel2pwm.txt");
    myIntervalMap.Integers("res/pwm2vel.txt");    
    
    for (c1=0;c1<myControl.length;c1++)
      myControl[c1]=new Control(VEL,0.4,0.1,0.1);
    for (c1=ELEF;c1<=ELEB;c1++) {
      myControl[c1].CN=VEL*0.5;
      myControl[c1].ei=1;
    }
    
    for (c1=0;c1<myRolk.length;c1++)
      myRolk[c1]=new Rolk();
    myRolk[ROLK_FRONT ].put(new Integer(0),new int[]{Rolk.MINI,Rolk.NULL,Rolk.PUSH,Rolk.NULL});
    myRolk[ROLK_FRONT ].put(new Integer(1),new int[]{Rolk.MAXI,Rolk.NULL,Rolk.PULL,Rolk.NULL});    
    myRolk[ROLK_FULL  ].put(new Integer(0),new int[]{Rolk.MINI,Rolk.MINI,Rolk.PUSH,Rolk.NULL});
    myRolk[ROLK_FULL  ].put(new Integer(1),new int[]{Rolk.MAXI,Rolk.MAXI,Rolk.PULL,Rolk.PUSH});
    myRolk[ROLK_FULL  ].put(new Integer(2),new int[]{Rolk.MAXI,Rolk.MINI,Rolk.NULL,Rolk.PUSH});
    myRolk[ROLK_REAR  ].put(new Integer(0),new int[]{Rolk.NULL,Rolk.MAXI,Rolk.NULL,Rolk.PULL});
    myRolk[ROLK_REAR  ].put(new Integer(1),new int[]{Rolk.NULL,Rolk.MINI,Rolk.NULL,Rolk.PUSH});
    myRolk[ROLK_WORM  ].put(new Integer(0),new int[]{Rolk.MINI,Rolk.MAXI,Rolk.PUSH,Rolk.PULL});
    myRolk[ROLK_WORM  ].put(new Integer(1),new int[]{Rolk.MAXI,Rolk.MINI,Rolk.PULL,Rolk.PUSH});    
    myRolk[ROLK_EXTEND].put(new Integer(0),new int[]{Rolk.MAXI,Rolk.MAXI,Rolk.PULL,Rolk.PUSH});
    myRolk[ROLK_EXTEND].put(new Integer(1),new int[]{Rolk.MINI,Rolk.MINI,Rolk.PULL,Rolk.PUSH});    
  }
  
  static void POSreset() {
    int c1;
    for (c1=0;c1<6;c1++)
      pos[c1]=0;
    pos[6]=-ANG;
    pos[7]= ANG;
    for (c1=0;c1<8;c1++) {
      enc[c1]=RADenc(pos[c1]);
      bcl[c1]=pos[c1];
    }
  }
  
  static double ICRfreq(int icr)                { return 8000000/(double)(icr+1);                               }  
  static double ADCgyr(int adc)                 { return 0.0170274*(YAW-adc);                                   }
  static double ADCvcc(int adc)                 { return 0.06768049*adc-41.0823;                                }  
  static double ADCrad(int idx,int adc)         { return -(TLA[idx]+TLB[idx]-2*adc)/(TLA[idx]-TLB[idx])*TLT;    }
  static double CONsec(int con)                 { return (con+1)/2000000.;                                      }  
  static double ENCrad(long enc)                { return enc/18302.81845;                                       }  
  static long   RADenc(double pos)              { return (long)(pos*18302.81845);                               }
  static double PWMvel_vcc(int ind,int pwm)     { return PWMvel(ind,pwm)*vcc/_VCC;                              }
  static double MLA6()                          { return Math.sqrt(LEV*LEV+MLA*MLA-2*LEV*MLA*Math.cos(bcl[6])); }  
  static double MLA7()                          { return Math.sqrt(LEV*LEV+MLA*MLA-2*LEV*MLA*Math.cos(bcl[7])); }
  
  static double PWMvel(int ind,int pwm) { // converts pwm to vel at 24 V
    if (0<=pwm)
      return  myIntervalMap.get( pwm)*BST[ind];
    else
      return -myIntervalMap.get(-pwm)*BST[ind];
  }
  
  static int VELpwm(int ind,double vel) { // converts vel to pwm at 24 V
    if (0<=vel)
      return  myIntervalMap.get( vel/BST[ind]);
    else                                
      return -myIntervalMap.get(-vel/BST[ind]);
  }
  
  static int VELpwm_vcc(int ind,double vel) { // converts vel to pwm at current vcc
    if (0<=vel)
      return  myIntervalMap.get( vel*_VCC/(BST[ind]*vcc));
    else                                      
      return -myIntervalMap.get(-vel*_VCC/(BST[ind]*vcc));
  }
  
  static void EIC(int ind,double dst) { // ind is either ELEF, ELEB
    int c1=ind-ELEF+6;
    double fac=0.5;
    double dir=myControl[PUSH].dd;
    if (joi<Math.abs(lod[c1])) { // if load on joint is high
      if (ind==ELEF)
        dst=0+dir; // axle of joint remains static
      else
        dst=1-dir;
      fac=1;
    }
    if (Math.abs(BCL(c1)*vel[c1])<0.1)
      myControl[ind].ei=dst;
    else
      if (myControl[ind].ei<dst)
        myControl[ind].ei=Math.min(    myControl[ind].ei+Control.dlt*fac,dst);
      else
        myControl[ind].ei=Math.max(dst,myControl[ind].ei-Control.dlt*fac    );    
  }
  
  static double BCL(int ind) { // return in 0,...,1 depending only on sign of speed vel
    int sgn=0;
    if (Control.clk<cal+1.0) return 1; // during calibration phase      
    if (0.006<vel[ind]       ) sgn= 1;
    if (      vel[ind]<-0.006) sgn=-1;
    if (0==sgn)
      sgn=(int)Math.signum(rat[ind]);
    if (0==sgn)
      return 0;
    if (0<sgn)
      return (pos[ind]-bcl[ind])/BCL[ind];
    else
      return (bcl[ind]+BCL[ind]-pos[ind])/BCL[ind];
  }
  
  static void THR() {
    int c1;
    double lds=0;
    double dts=0;
    double dmx=0.1;
    double tmp;
    for (c1=0;c1<8;c1++) {
      lds+=lod[c1]*lod[c1];
      tmp=Math.abs(pwm[c1])/(double)icr[c1];
      dmx=Math.max(dmx,tmp);
      dts+=tmp;
    }
    double fac=dts/dmx;
    thr=Math.min((UPS-lds*XIS[1])/(XIS[0]*fac),1);    
  }
  
  static double SLA() { // estimation of slope [CHECKED:28 Jul]
    double ret,c;
    ret=Math.acos(grv[0])-1.57079632;
    c=MLA6();
    ret-=Math.acos((MLA*MLA+c*c-LEV*LEV)/(2*MLA*c))*.5;
    c=MLA7();
    ret+=Math.acos((MLA*MLA+c*c-LEV*LEV)/(2*MLA*c))*.5;
    return ret;
  }
  
  static void FLDbuild() {
    int c1,c2;
  	double a=LEV;
    double b=MLA;
    double r=RAD;
    double g,c,u;
    // build PUSH vector field
    fld[PUSH][0]= 1;
    fld[PUSH][1]= 1;
    fld[PUSH][2]= 1;
    fld[PUSH][3]= 1;
    fld[PUSH][4]= 1;
    fld[PUSH][5]= 1;
    fld[PUSH][6]= 0;
    fld[PUSH][7]= 0;
    // build TURN vector field
    fld[TURN][0]= 5.3*MLA6()+0.04;
    fld[TURN][1]=-fld[TURN][0];
    fld[TURN][2]= 1;
    fld[TURN][3]=-1;
    fld[TURN][4]= 5.3*MLA7()+0.04;
    fld[TURN][5]=-fld[TURN][4];
    fld[TURN][6]= 0;
    fld[TURN][7]= 0;
    g=1/Math.max(fld[TURN][0],fld[TURN][4]); // either front or back axis are faster
    for (c1=0;c1<8;c1++)
      fld[TURN][c1]*=g;
    // build DIGG vector field
    fld[DIGG][0]= 0;
    fld[DIGG][1]= 1;
    fld[DIGG][2]= 0;
    fld[DIGG][3]= 0;
    fld[DIGG][4]= 0;
    fld[DIGG][5]= 0;
    fld[DIGG][6]= 0;
    fld[DIGG][7]= 0;
    // build DRFT vector field
    fld[DRFT][0]=-1;
    fld[DRFT][1]= 1;
    fld[DRFT][2]= 0;
    fld[DRFT][3]= 0;
    fld[DRFT][4]= 1;
    fld[DRFT][5]=-1;
    fld[DRFT][6]= 0;
    fld[DRFT][7]= 0;
    // build ELEF vector field
    fld[ELEF][0]=0;
    fld[ELEF][1]=0;
    fld[ELEF][2]=0;
    fld[ELEF][3]=0;
    fld[ELEF][4]=0;
    fld[ELEF][5]=0;
    fld[ELEF][6]=1;
    fld[ELEF][7]=0;
    // build ELEB vector field
    fld[ELEB][0]=0;
    fld[ELEB][1]=0;
    fld[ELEB][2]=0;
    fld[ELEB][3]=0;
    fld[ELEB][4]=0;
    fld[ELEB][5]=0;
    fld[ELEB][6]=0;
    fld[ELEB][7]=1;
    // build ELEF passive vector field 
    c=MLA6();    
    u=a*b*Math.sin(-bcl[6])/(r*c); // axis motion according to eich
	  fle[ELEF][0]=-u*   myControl[ELEF].ei;
	  fle[ELEF][1]=      fle[ELEF][0];
	  fle[ELEF][2]= u*(1-myControl[ELEF].ei);
	  fle[ELEF][3]=      fle[ELEF][2];
	  fle[ELEF][4]=      fle[ELEF][2];
	  fle[ELEF][5]=      fle[ELEF][2];
    fle[ELEF][6]=0;
    fle[ELEF][7]=0;    
    u=b*(b-a*Math.cos(-bcl[6]))/(c*c); // compensation for suspension tilt
    fle[ELEF][0]-=u;
    fle[ELEF][1]-=u;
    u=0.5*a*(a-b*Math.cos(-bcl[6]))/(c*c);
    fle[ELEF][2]+=u;
    fle[ELEF][3]+=u;
    // build ELEB passive vector field
    c=MLA7();    
    u=a*b*Math.sin(+bcl[7])/(r*c); // axis motion according to eich
	  fle[ELEB][0]= u*(1-myControl[ELEB].ei);
	  fle[ELEB][1]=      fle[ELEB][0];
	  fle[ELEB][2]=      fle[ELEB][0];
	  fle[ELEB][3]=      fle[ELEB][0];
	  fle[ELEB][4]=-u*   myControl[ELEB].ei;
	  fle[ELEB][5]=      fle[ELEB][4];
    fle[ELEB][6]=0;
    fle[ELEB][7]=0;
    u=b*(b-a*Math.cos(+bcl[7]))/(c*c); // compensation for suspension tilt
    fle[ELEB][4]-=u;
    fle[ELEB][5]-=u;
    u=0.5*a*(a-b*Math.cos(+bcl[7]))/(c*c);
    fle[ELEB][2]+=u;
    fle[ELEB][3]+=u;
  }
  
  static private double[] clp_a=new double[8];
  static private double[] clp_b=new double[8];
  static private double[] clp_c=new double[8];
  
  static void FLDrat() {
    int c1,c2;
	  double elef=BCL(6);
	  double eleb=BCL(7);
	  double lam=0.9;
	  if (Control.clk<cal+1.0) lam=0; // during calibration phase	    
	  kap=1;
    for (c1=0;c1<8;c1++) {
      clp_a[c1]=0;
      for (c2=PUSH;c2<=DRFT;c2++)
        clp_a[c1]+=fld[c2][c1]*myControl[c2].vn;
      clp_b[c1]=0;
      for (c2=ELEF;c2<=ELEB;c2++)
        clp_b[c1]+=fld[c2][c1]*myControl[c2].vn;
	    clp_b[c1]+=fle[ELEF][c1]*elef*(vel[6]*lam+myControl[ELEF].vn*(1-lam));
 	    clp_b[c1]+=fle[ELEB][c1]*eleb*(vel[7]*lam+myControl[ELEB].vn*(1-lam));
 	    clp_c[c1]=PWMvel_vcc(c1,(int)Math.round(thr*icr[c1]));
      if (clp_b[c1]<=clp_c[c1]) {
        if (clp_a[c1]!=0)
          if (0<clp_a[c1]*clp_b[c1])
            kap=Math.min(kap,(clp_c[c1]-Math.abs(clp_b[c1]))/Math.abs(clp_a[c1]));
          else
            kap=Math.min(kap,(clp_c[c1]+Math.abs(clp_b[c1]))/Math.abs(clp_a[c1]));
      } else 
        kap=0;
    }
    kap=Math.max(kap,0); 
    for (c1=0;c1<8;c1++)
      rat[c1]=clp_a[c1]*kap+clp_b[c1];
    for (c2=PUSH;c2<=DRFT;c2++) // truncate instantaneous control
      myControl[c2].vn*=kap;
  }
  
  static void LODlod() {
    int c1,c2;
    for (c1=0;c1<8;c1++) {
      VEX.push(new Double(vex[c1]));
      lod[c1]=vel[c1];
    }
    for (c2=0;c2<LOD.length;c2++)
      for (c1=0;c1<8;c1++)
        lod[c1]-=LOD[c2]*((Double)VEX.pop()).doubleValue();
    // begin: compute mean and variance
    lod_mea=0;
    for (c1=0;c1<6;c1++)
      lod_mea+=lod[c1];
    lod_mea*=0.166667;
    lod_var=0;
    for (c1=0;c1<6;c1++)
      lod_var+=(lod[c1]-lod_mea)*(lod[c1]-lod_mea);
    lod_var*=0.166667;
    // end: compute mean and variance
  }
  
  static void RAGE(double eps) {
    double yf=LEV*MLA*Math.sin(-bcl[6])/(RAD*MLA6())*BCL(6)*vel[6];
    double yr=LEV*MLA*Math.sin(+bcl[7])/(RAD*MLA7())*BCL(7)*vel[7];
    double den=2*(yf*yf+yr*yr);
    double hf,hr;
    if (0.01<den) {
      hf=(yf*yf+(2*eps-1)*yf*yr+2*eps*yr*yr)/den;
      hr=(2*(1-eps)*yf*yf+(1-2*eps)*yf*yr+yr*yr)/den;
      EIC(ELEF,hf);
      EIC(ELEB,hr);
    }
  }
  
  static void ACCdigest() {
    int c1;
    double len;
    for (c1=0;c1<3;c1++)
      acc[c1]=(ACC[c1][0]+ACC[c1][1]-2*acc[c1])/(ACC[c1][0]-ACC[c1][1]);
    len=0;
    for (c1=0;c1<3;c1++)
      len+=acc[c1]*acc[c1];
    len=0.2/Math.sqrt(len);
    for (c1=0;c1<3;c1++)
      grv[c1]+=acc[c1]*len;
    len=0;
    for (c1=0;c1<3;c1++)
      len+=grv[c1]*grv[c1];
    len=1/Math.sqrt(len);
    for (c1=0;c1<3;c1++)
      grv[c1]=grv[c1]*len;
  }
  
  static public void PWMvex() {
    int c1;
    for (c1=0;c1<8;c1++)
      vex[c1]=PWMvel_vcc(c1,pwm[c1]);
  }
  
	static public void run() throws IOException {
  	int c1,c2;
  	double tmp,lam;
	  for (c1=0;c1<8;c1++) 
	    ral[c1]=0;  	
    if ((logic&LOGIC_CALIBRATION)==LOGIC_CALIBRATION&&calib!=CALIB_PASSIVE)
      cal=Control.clk;
    if (Control.clk<cal+1.0) {
      led|=LED_CALIB;
      logic&=LOGIC_CALIBRATION; // reset logic
      for (c1=PUSH;c1<=ELEB;c1++)
        myControl[c1].vn=0;
      myControl[ELEF].ei=1;
      myControl[ELEB].ei=1;
      if ((logic&LOGIC_CALIBRATION)==LOGIC_CALIBRATION) {
        if (CALIB_PASSIVE<calib)
          fuse=0;
        switch (calib) {
          case CALIB_ELEF_UP: myControl[ELEF].vn=-CAL; break;
          case CALIB_ELEF_DN: myControl[ELEF].vn= CAL; break;
          case CALIB_ELEB_UP: myControl[ELEB].vn= CAL; break;
          case CALIB_ELEB_DN: myControl[ELEB].vn=-CAL; break;
          case CALIB_IGNITE:
            ral[6]=-0.16;
            ral[7]=-0.16;
            fuse=1; // finally defuse
            break;
        }
      }
      POSreset();
    } else {
      led&=~LED_CALIB;
      for (c1=PUSH;c1<=ELEB;c1++)       // dead man's switch
        myControl[c1].iVD();            // decelerate motors after communication timeout
        
      if (!Rolk.isActive()) {
        if ((logic&LOGIC_SMART_PLACE)==LOGIC_SMART_PLACE) {
          tmp=myControl[PUSH].dd;
          if (myControl[ELEF].vd<0) EIC(ELEF,1-tmp);
          if (myControl[ELEF].vd>0) EIC(ELEF,0+tmp);
          if (myControl[ELEB].vd>0) EIC(ELEB,0+tmp);
          if (myControl[ELEB].vd<0) EIC(ELEB,1-tmp);
        } else {
          if (0.01<Math.abs(myControl[ELEF].vd)) EIC(ELEF,1);
          if (0.01<Math.abs(myControl[ELEB].vd)) EIC(ELEB,1);
        }
      }
      if ((logic&LOGIC_TURN_CONTRACT)==LOGIC_TURN_CONTRACT) {
        double tcn=Math.min(0.5*Math.abs(myControl[TURN].vn),VEL*0.5);
        if (0.1<tcn) {
          myControl[ELEF].vd+=tcn;
          myControl[ELEB].vd-=tcn;
          EIC(ELEF,.666);
          EIC(ELEB,.666);
        }
      }
      myRolk[Rolk.mode].V();
      for (c1=PUSH;c1<=DRFT;c1++)
        myControl[c1].V();
      for (c1=ELEF;c1<=ELEB;c1++) {
        c2=c1-ELEF+6;
        myControl[c1].V(bot[c2],top[c2]);
      }
      myControl[ELEF].vn*=fuse;
      myControl[ELEB].vn*=fuse;
	  }
	  
	  FLDrat();

	  if (0.05<Math.abs(myControl[PUSH].vn)&&(logic&LOGIC_LOAD_DISTRIB_V)==LOGIC_LOAD_DISTRIB_V)
  	  for (c1=0;c1<6;c1++)
  	    ral[c1]=Control.clip(-0.3,5*(lod_mea-lod[c1])*lod_var,0.3);
    for (c1=0;c1<8;c1++)
      pwm[c1]=VELpwm_vcc(c1,rat[c1]+ral[c1]);

    bot[6]=-ANG-EXP;
    top[6]=-ANG;
    bot[7]= ANG;
    top[7]= ANG+EXP;

    for (c1=0;c1<8;c1++)
      if (bcl[c1]<bot[c1]-TOL||top[c1]+TOL<bcl[c1])
        pwm[c1]=0;

    PWMvex();
      
    if (dirty)
      Mediator.sendCfg();
  	Mediator.read();
    Mediator.send();
    
	  boolean stopped=true;
	  for (c1=0;c1<8;c1++)
	    stopped&=(pwm[c1]==0);
	  lam=0.95;
	  if (stopped) {
	    YAW=         lam*YAW+(1-lam)*yaw;
	    VCC=Math.max(lam*VCC+(1-lam)*vcc,vcc);
    } else
      VCC=Math.max(VCC,vcc);

    dirty=false;
    
    if ((logic&LOGIC_SLOPE_RESPONSE)==LOGIC_SLOPE_RESPONSE) {
      double sla=SLA();
      double sin=Math.sin(Math.abs(sla));
      double oms=sin*sin-1;
      double mlo=MLA*LEV*oms;
      double ml2=MLA*MLA;
      double gam=Math.acos( (Math.sqrt( mlo*mlo*ml2*sin*sin * (ml2+LEV*LEV*oms) )-mlo*mlo ) / (mlo*ml2) );
      if (0<sla) {
        bot[7]= gam;
        if (!myControl[ELEB].bP(bot[7],top[7])) // bounds are violated
          EIC(ELEB,1);                          // advance joint axle
      } else {
        top[6]=-gam;
        if (!myControl[ELEF].bP(bot[6],top[6])) // bounds are violated
          EIC(ELEF,1);                          // advance joint axle 
      }
    }

    LODlod(); // computes load on the wheels (when here, con=1 => 8*6 ringbuffer)
    THR();    
    
    for (c1=0;c1<8;c1++) {
      if (bcl[c1]+BCL[c1]<pos[c1])
        bcl[c1]=pos[c1]-BCL[c1];
      if (pos[c1]<bcl[c1])
        bcl[c1]=pos[c1];
    }
    
    FLDbuild();
    
	  for (c1=PUSH;c1<=ELEB;c1++)
	    myControl[c1].vd=0;
    for (c1=ELEF;c1<=ELEB;c1++) {
      myControl[c1].pn=bcl[6+c1-ELEF];  	  
  	  myControl[c1].pd=bcl[6+c1-ELEF];
	  }      
  }
}
