// code by amaula

#ifndef _CONNECTION_H_
#define _CONNECTION_H_

#define CONNECTION_OK             0
#define CONNECTION_ERROR          -1

typedef int (*sendDataFuncPtr)    (void* d, const unsigned char* buffer, int bufLen);
typedef int (*receiveDataFuncPtr) (void* d, unsigned char* buffer, int bufLen);
typedef int (*receiveTimedDataFuncPtr) (void* d, unsigned char* buffer, int bufLen, int timeout_ms);
typedef int (*connectFuncPtr)     (void* d );
typedef int (*disconnectFuncPtr)  (void* d );
typedef int (*getHandleFuncPtr)   (void* d );
typedef int (*getUnsentBytesFuncPtr) (void* d );
typedef void (*deinitFuncPtr)     (void* d );
typedef const char* (*getPeerIPFuncPtr) (void* d);
typedef int (*getPeerPortFuncPtr) (void* d);

typedef struct _connectionPtr {
  void*                           dataPtr;
  sendDataFuncPtr                 sendFunc;
  receiveDataFuncPtr              receiveFunc;
  receiveTimedDataFuncPtr         receiveTimedFunc;
  connectFuncPtr                  connectFunc;
  disconnectFuncPtr               disconnectFunc;
  getHandleFuncPtr                getHandleFunc;
  getUnsentBytesFuncPtr           getUnsentBytesFunc;
  deinitFuncPtr                   deinitFunc;
  getPeerPortFuncPtr              getPeerPortFunc;
  getPeerIPFuncPtr                getPeerIPFunc;
} TConnectionPtr;

typedef struct _connection {
  const TConnectionPtr* link;     //< Associated link to protocol (OWNED HERE)
} TConnection;


// Calling this Allocates and inits
// @param con        Pointer to connection parameters
TConnection* connectionConstruct(const TConnectionPtr* cp);

// Calling this Deinits & Destroys
// @param con        Pointer to connection parameters
void connectionDestroy(TConnection* con);

// Calling this calls the modules CONNECT function.
// @param con        Pointer to connection parameters
int connectionConnect(TConnection* con);

// Calling this calls the modules DISCONNECT function.
// @param con        Pointer to connection parameters
int connectionDisconnect(TConnection* con);

// @param con        Pointer to connection parameters
const void* connectionGetDataPtr(const TConnection* con);

// @param con        Pointer to connection parameters
int connectionGetHandle(const TConnection* con);

// Returns the number of bytes written to pipe, but not sent to
// network yet.
// @param con        Pointer to connection parameters
int connectionGetUnsentBytes(const TConnection* con);

// This initializes the Connection modules to use module 'cp' as
// implementation.
// @param con        Pointer to connection parameters
int connectionInit(TConnection* con,const TConnectionPtr* cp);

// @param con        Pointer to connection parameters
int connectionDeinit(TConnection* con);

// Function that sends raw data over the connection.
// Connection must be opened with Connect() before this
// function can be called.
// The function takes care of connection state checking, so
// it's always safe to call this function.
// @param con        Pointer to connection parameters
// @param buffer     Pointer to data buffer to send
// @param bufLen     Number of bytes to send
// @return           Real number of bytes sent, or negative on error.
int connectionSendData(TConnection* con,const unsigned char* buffer,int bufLen);

// Function that receives raw data over the connection.
// Connection must be opened with Connect() before this
// function can be called.
// The function takes care of connection state checking, so
// it's always safe to call this function.
// @param con        Pointer to connection parameters
// @param buffer     Pointer to data buffer to save received data to.
// @param bufLen     MAXIMUM number of bytes to receive
// @return           Real number of bytes received, or negative on error.
int connectionReceiveData(TConnection* con,unsigned char* buffer,int bufLen);

int connectionReceiveDataTimed(TConnection* con,unsigned char* buffer,int bufLen,int timeout_ms);

const char* connectionGetPeerIP(const TConnection* con);
int connectionGetPeerPort(const TConnection* con);

#endif
